--[[
# Copyright 2020 Quarq
# This file is part of A-Bomb: Azeroth Bombardier.
# A-Bomb: Azeroth Bombardier is distributed under a BSD License.
# It is provided AS-IS and all warranties, express or implied, including, but not
# limited to, the implied warranties of merchantability or fitness for a particular
# purpose, are disclaimed.  See the LICENSE file for full information.
--]]
local MYNAME, _ = ...
MYNAME = GetAddOnMetadata(MYNAME, "X-ShortName") or MYNAME
local Addon = _G[MYNAME]
assert(Addon)


-- states
local FALLING, EXPLODING, FINISHED = "FALLING", "EXPLODING", "FINISHED"


local THROW_SPEED = 0.075	-- bomb movement speed in map units per second

local FUSES = { 702, 3244, 3347, 6437 }	-- sound IDs

local EXPLOSIONS = { 4100, 4775, 6522 } -- sound IDs
local EXPLOSION_DURATION = 0.50	-- seconds



--
--  Map Marker / Data Provider / Pin Pool implementation
--

-- my marker
local Prototype = CreateFromMixins(MapCanvasPinMixin)
local POOLNAME = "ABombBomb" -- bomb-a-loo-bomb ABomb-bam-boom


local _provider = CreateFromMixins(MapCanvasDataProviderMixin)
WorldMapFrame:AddDataProvider(_provider)

function _provider : OnMapChanged()
	-- local mapid = self:GetMap():GetMapID()
    -- print("map changed to", mapid)
end


local _pool = CreateFramePool("FRAME")
WorldMapFrame.pinPools[POOLNAME] = _pool

function _pool.creationFunc(pool)
    return Mixin(CreateFrame("FRAME"), Prototype)
end

function _pool.resetterFunc(pool, pin)
    FramePool_HideAndClearAnchors(pool, pin)
    pin:OnReleased()
	-- these were assigned by MapCanvas:AcquirePin
    pin.pinTemplate = nil
    pin.owningMap = nil
end




function Prototype : Reset()
	self.state = nil
	self.remaining = nil

	self.detonated = nil
	self.finished = nil

	if (self.texture) then
		self.texture:SetRotation(0)
	end

	if (self.combo) then
		if (self.combo_timer) then
			Addon.TIMER:Cancel(self.combo_timer)
			self.combo_timer = nil
		end
		self.combo:Hide()
		self.combo:SetAlpha(1)
	end
end


function Prototype : PlaySoundForState(state)
	if (self.sound) then
		StopSound(self.sound)
		self.sound = nil
	end
	if (state == FALLING) then
		_, self.sound = PlaySound(3244)		-- 3244 is a campfire burning
		-- _, self.sound = PlaySound(3257)	-- 3257 is a flare sound (too much)
		-- _, self.sound = PlaySound(6763)	-- sizzle (too hissy)
	elseif (state == EXPLODING) then
		local which = math.random(1, #EXPLOSIONS)
		_, self.sound = PlaySound(EXPLOSIONS[which])
	end
end


function Prototype : StartFalling()
	assert(self.state == nil)

	self:Reset()

	self.state = FALLING

	local dx, dy = self.tx-self.px, self.ty-self.py
	self.dist = math.sqrt(dx*dx + dy*dy)
	self.vx = THROW_SPEED * dx/self.dist		-- x/y components of bomb's flight, in map units per second
	self.vy = THROW_SPEED * dy/self.dist

	self.texture:SetTexture("Interface/Addons/A-Bomb/textures/bomb")
	self:SetSize(10,10)
	self:SetAlpha(1)

	self.remaining = self.dist / THROW_SPEED	-- the travel time, in seconds

	self:PlaySoundForState(self.state)

	self.elapsed = 0
	self:SetScript("OnUpdate", self.OnUpdate)
end


function Prototype : StartExploding()
	assert(self.state == FALLING)

	self:Reset()

	self.state = EXPLODING
	self.detonated = true

	self.texture:SetTexture("Interface/Addons/A-Bomb/textures/explosion")
	self:SetSize(10,10)
	self:SetAlpha(1)

	self.remaining = EXPLOSION_DURATION

	self:PlaySoundForState(self.state)
	self.sound = nil -- boom sound will stop on its own
end


function Prototype : ShowCombo(count)
	if (not self.combo) then
		self.combo = self:CreateFontString(nil, "OVERLAY", nil, 7)
		self.combo:SetFont("Fonts/FRIZQT__.TTF", 18)
		self.combo:SetTextColor(1,1,1, 1)
		self.combo:SetShadowOffset(1,-1)
		self.combo:SetShadowColor(1,1,1, 0.50)
	end
	self.combo:SetText(""..count)
	self.combo:ClearAllPoints()
	local y = 8
	self.combo:SetPoint("CENTER", self, "CENTER", 0, y)
	self.combo:SetAlpha(1)
	self.combo:Show()
	local bomb = self
	self.combo_timer = Addon.TIMER:Start(EXPLOSION_DURATION,
	                                     function(timer) y=y+0.2; bomb.combo:SetPoint("CENTER", bomb, "CENTER", 0, y) end,
										 function() bomb.combo_timer=nil end)
end


function Prototype : Finish()
	assert(self.state == EXPLODING)

	self:Reset()

	self.state = FINISHED
	self.finished = true

	self.texture:SetTexture(nil)

	self:SetScript("OnUpdate", nil)
end


function Prototype : Abort()
	self:Reset()
	self:SetScript("OnUpdate", nil)
	if (self.sound) then
		StopSound(self.sound)
		self.sound = nil
	end
end


function Prototype : OnAcquired(px,py, tx,ty)

	self:SetPosition(px,py)

	self.px, self.py = px, py
	self.tx, self.ty = tx, ty

	if (not self.texture) then
		self.texture = self:CreateTexture(nil, "OVERLAY", nil, 0)
		self.texture:SetAllPoints(self)
	end

	self:StartFalling()

    self:UseFrameLevelType("PIN_FRAME_LEVEL_AREA_POI")
end


function Prototype : OnReleased()
	self:Reset()
	self:SetScript("OnUpdate", nil)
end



function Prototype : OnUpdate(dt)
	self.elapsed = self.elapsed + dt

	if (self.state == FALLING) then
		self.remaining = self.remaining - dt

		self.texture:SetRotation(self.elapsed * 4*math.pi)

		local dx = self.vx * self.elapsed
		local dy = self.vy * self.elapsed
		local fallen = math.sqrt(dx*dx + dy*dy)
		if (fallen < self.dist) then
			-- advance its fall
			self:SetPosition(self.px + dx, self.py + dy)
		else
			-- position at the target
			self:SetPosition(self.tx, self.ty)

			self:StartExploding()
			assert(self.state == EXPLODING)
			assert(self.detonated)
		end

	elseif (self.state == EXPLODING) then
		self.remaining = self.remaining - dt

		if (self.remaining > 0) then
			local scale = 1.0 + 2.0 - 1.5*self.remaining/EXPLOSION_DURATION
			self:SetSize(8*scale, 8*scale)

			local alpha = self.remaining>EXPLOSION_DURATION*3/4 and 1.0 or self.remaining/(EXPLOSION_DURATION*3/4)
			self:SetAlpha(alpha)
		else
			self:Finish()
			assert(self.state == FINISHED)
			Addon:RemoveBomb(self)
		end
	else
		assert(false) -- shouldn't be handling OnUpdate in any other state
	end
end


function Addon : DrawBomb(px,py, tx,ty)
 	return _provider:GetMap():AcquirePin(POOLNAME, px,py, tx,ty)
end

function Addon : RemoveBomb(bomb)
	_provider:GetMap():RemovePin(bomb)
end
