--[[
# Copyright 2020 Quarq
# This file is part of A-Bomb: Azeroth Bombardier.
# A-Bomb: Azeroth Bombardier is distributed under a BSD License.
# It is provided AS-IS and all warranties, express or implied, including, but not
# limited to, the implied warranties of merchantability or fitness for a particular
# purpose, are disclaimed.  See the LICENSE file for full information.
--]]
local MYNAME, _ = ...
MYNAME = GetAddOnMetadata(MYNAME, "X-ShortName") or MYNAME
local Addon = _G[MYNAME]
assert(Addon)

-- grab these before Questie "cleans them up" by deleting them
local QuestieDB = QuestieLoader:ImportModule("QuestieDB")
local QuestieData = {
	questData=QuestieDB.questData, questKeys=QuestieDB.questKeys,
	npcData=QuestieDB.npcData, npcKeys=QuestieDB.npcKeys,
	objectData=QuestieDB.objectData, objectKeys=QuestieDB.objectKeys,
	itemData=QuestieDB.itemData, itemKeys=QuestieDB.itemKeys,
}
assert(QuestieData.npcData)


local QuestieZoneDB = QuestieLoader:ImportModule("ZoneDB")

local _mapid2area, _area2mapid = {}, {}
for area, mapid in pairs(QuestieZoneDB.private.areaIdToUiMapId) do
	_mapid2area[mapid] = area
	_area2mapid[area] = mapid
end

local function AreaForZoneID(mapid)
	return _mapid2area[mapid]
end

local function ZoneForAreaID(areaid)
	return _area2mapid[areaid]
end


local floor, sqrt = math.floor, math.sqrt
local sprintf = string.format


local INDEX = {}
Addon.INDEX = INDEX
local GW, GH = Addon.GW, Addon.GH

--[[
INDEX[zone] = table of grids
grids[key] = array of mobs in this grid, each like { npcid, zone, x, y }
--]]

function Addon : BuildIndex(zone)
	assert(zone)
	local area = AreaForZoneID(zone)
	if (area == nil) then return end

	local grids = INDEX[zone] or {}
	INDEX[zone] = grids

	local keys = QuestieData.npcKeys

	local minLevel = 2
	local maxLevel = 999

	local playerFaction = UnitFactionGroup("player")
	local otherFaction = (playerFaction == "Alliance") and "H" or "A"

	local t = debugprofilestop()
	local n = 0
	local npc = {}
	for npcid, data in pairs(QuestieData.npcData) do
		setmetatable(npc, { __index = function(t,k) return data[keys[k]] end })
		if (npc.minLevel == nil or npc.maxLevel == nil) then
			-- skip
		elseif (npc.minLevel < minLevel or npc.maxLevel > maxLevel) then
			-- skip
		elseif (npc.rank and npc.rank > 0) then
			-- skip
		elseif (npc.friendlyToFaction == nil or npc.friendlyToFaction == otherFaction) then -- friendly to Horde, or to noone
			if (npc.spawns and npc.spawns[area]) then
				for _, p in pairs(npc.spawns[area]) do
					local x,y = p[1]/100, p[2]/100
					local gridkey = sprintf("%02d:%02d", floor(x/GW), floor(y/GH))

					local mobs = grids[gridkey] or {}
					grids[gridkey] = mobs

					tinsert(mobs, { npcid=npcid, zone=zone, x=x, y=y } )
					n = n + 1
				end
			end
		end
	end
	t = debugprofilestop() - t
	-- print("zone", zone, "with", n, "points took", t, "ms")

	return (n>0 and grids or nil)
end

--[[
returns simple array of mobs, each like {npcid, zone, x, y}
--]]

function Addon : GetMobsNear(zone, x, y)
	local grids = INDEX[zone]
	if (grids == nil) then
		grids = self:BuildIndex(zone)
		if (grids == nil) then
			return {}
		end
	end
	assert(grids)

	local mobs = {}
	local n = 0

	local gx, gy = floor(x/GW), floor(y/GH)
	for dx = -1, 1 do
		for dy = -1, 1 do
			local key = sprintf("%02d:%02d", gx+dx, gy+dy)
			if (grids[key]) then
				for _, mob in pairs(grids[key]) do
					tinsert(mobs, mob)
					n = n + 1
				end
			end
		end
	end
	-- print(zone, "("..x..", "..y..")", "has", n, "mobs nearby")
	return mobs
end



--
--  Map Marker / Data Provider / Pin Pool implementation
--

-- my marker
local Target = CreateFromMixins(MapCanvasPinMixin)
local POOLNAME = "ABombTarget"


local _provider = CreateFromMixins(MapCanvasDataProviderMixin)
WorldMapFrame:AddDataProvider(_provider)

function _provider : OnMapChanged()
	-- local mapid = self:GetMap():GetMapID()
    -- print("map changed to", mapid)
end


local _pool = CreateFramePool("FRAME")
WorldMapFrame.pinPools[POOLNAME] = _pool

function _pool.creationFunc(pool)
    return Mixin(CreateFrame("FRAME"), Target)
end

function _pool.resetterFunc(pool, pin)
    FramePool_HideAndClearAnchors(pool, pin)
    pin:OnReleased()
	-- these were assigned by MapCanvas:AcquirePin
    pin.pinTemplate = nil
    pin.owningMap = nil
end


function Target : OnAcquired(x, y, zone, color)

	self:SetPosition(x, y)
	self:SetSize(5,5)

	self.zone = zone -- the zone the mob is in, not necessarily the zone currently mapped (to which x,y position refers)
	self.color = color

	self.dead = nil

	if (not self.texture) then
		self.texture = self:CreateTexture(nil, "OVERLAY", nil, 0)
		self.texture:SetAllPoints(self)
	end

	self.texture:SetTexture("Interface/Addons/A-Bomb/textures/target")

	local r,g,b = Addon.hsv(self.color.hue, 1, 1)
	self.texture:SetVertexColor(r,g,b, 1)

    self:UseFrameLevelType("PIN_FRAME_LEVEL_AREA_POI")
end


function Target : Die()
	assert(not self.dead)
	self:SetSize(11,11)
	self.texture:SetTexture("Interface/Addons/A-Bomb/textures/dead")
	self.texture:SetVertexColor(1,1,1, 1)
	self.dead = true

	-- print("killed", self.color:designator(7), "target")
end


function Target : OnReleased()
	-- nothing
end


function Addon : DrawTarget(x, y, ...)
 	return _provider:GetMap():AcquirePin(POOLNAME, x, y, ...)
end

function Addon : RemoveTarget(target)
 	return _provider:GetMap():RemovePin(target)
end

function Addon : RemoveAllTargets()
	_provider:GetMap():RemoveAllPinsByTemplate(POOLNAME)
	self.targets = {}
	self.ntargets = 0
end
